/**
 *
 * Project: Eight Track, File: GuildMusicHandler.java
 *
 * Copyright 2019-2019 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Eight-Track(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.discord.eighttrack.handlers;

import java.awt.Color;
import java.io.IOException;
import java.math.BigInteger;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;

import com.sedmelluq.discord.lavaplayer.player.AudioPlayer;
import com.sedmelluq.discord.lavaplayer.player.DefaultAudioPlayerManager;
import com.sedmelluq.discord.lavaplayer.source.AudioSourceManagers;
import com.sedmelluq.discord.lavaplayer.track.AudioTrack;

import net.dv8tion.jda.api.EmbedBuilder;
import net.dv8tion.jda.api.Permission;
import net.dv8tion.jda.api.entities.ChannelType;
import net.dv8tion.jda.api.entities.GuildChannel;
import net.dv8tion.jda.api.entities.Message.Attachment;
import net.dv8tion.jda.api.entities.MessageEmbed;
import net.dv8tion.jda.api.entities.VoiceChannel;
import net.dv8tion.jda.api.events.message.guild.GuildMessageReceivedEvent;
import net.dv8tion.jda.api.managers.AudioManager;
import net.locusworks.crypto.utils.HashUtils;
import net.locusworks.discord.eighttrack.database.entities.DiscordGuild;
import net.locusworks.discord.eighttrack.database.entities.GuildPlaylist;
import net.locusworks.discord.eighttrack.database.entities.GuildPlaylistSong;
import net.locusworks.discord.eighttrack.database.entities.GuildSong;
import net.locusworks.discord.eighttrack.database.entities.Song;
import net.locusworks.discord.eighttrack.scheduler.TrackScheduler;
import net.locusworks.discord.eighttrack.services.GuildSongRepoService;
import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.ApplicationLoggerFactory;

public class GuildMusicHandler {

  private Path musicDir;
  private TrackScheduler ts;
  private DefaultAudioPlayerManager apm;
  private ApplicationLogger logger;
  private AudioPlayer player;
  private AtomicBoolean playing;
  private Long voiceChannelId;

  private OffsetDateTime lastPlayed;

  private Consumer<Long> callback;
  private Mp3UploadHandler uploadHandler;
  private GuildSongRepoService guildSongRepoService;
  private long guildId;

  public GuildMusicHandler(Path musicDir, long guildId, Mp3UploadHandler uploadHandler, GuildSongRepoService guildSongRepoService) throws IOException {
    this.logger = ApplicationLoggerFactory.getLogger(GuildMusicHandler.class);
    this.playing = new AtomicBoolean(false);
    this.musicDir = musicDir;
    this.lastPlayed = OffsetDateTime.now();
    this.uploadHandler = uploadHandler;
    this.guildSongRepoService = guildSongRepoService;
    this.guildId = guildId;
    this.apm = new DefaultAudioPlayerManager();

    AudioSourceManagers.registerLocalSource(apm);
    this.player = apm.createPlayer();

    this.ts = new TrackScheduler();
    player.addListener(ts);
  }

  public Long getCurrentVoiceChannelId() {
    return voiceChannelId;
  }

  public OffsetDateTime getLastPlayed() {
    return lastPlayed;
  }

  public void accept(Consumer<Long> callback) {
    this.callback = callback;
  }

  public void isPlaying(boolean playing) {
    this.playing.set(playing);
  }

  public void upNext(GuildMessageReceivedEvent event) throws Exception {
    upNext(event, null);
  }

  public void upNext(GuildMessageReceivedEvent event, List<String> commands) throws Exception {
    if (commands != null && !commands.isEmpty()) {
      String uuid = commands.remove(0);
      boolean foundSong = findSong(uuid);
      if (!foundSong) {
        event.getChannel().sendMessageFormat("Unable to find song with identifier: %s", uuid).queue();
        return;
      }
    }
    if (ts.peek() == null) {
      loadRandomSong();
    }
    AudioTrack track = ts.peek();
    MessageEmbed embed = new EmbedBuilder()
        .setAuthor(event.getMember().getEffectiveName(), null, event.getAuthor().getAvatarUrl())
        .setTitle("Next Up:")
        .setDescription(String.format("**%s** by __%s__", track.getInfo().title, track.getInfo().author))
        .setTimestamp(OffsetDateTime.now())
        .build();
    event.getChannel().sendMessage(embed).queue();
  }

  public void repeat(GuildMessageReceivedEvent event) throws Exception {
    next(event);
    ts.setFinishedCallback((ater) -> {
      if (!playing.get()) return;
      try {
        next(event);
      } catch (Exception e) {
        logger.error(e);
      }
    });
  }

  public void next(GuildMessageReceivedEvent event) throws Exception {
    next(event, null);
  }

  public void next(GuildMessageReceivedEvent event, List<String> commands) throws Exception {
    if (!playing.get()) {
      play(event, commands);
      playing.set(true);
      return;
    }
    GuildChannel gc = event.getGuild().getGuildChannelById(ChannelType.VOICE, voiceChannelId);
    if (gc != null && gc.getMembers().size() == 1) {
      event.getChannel().sendMessage("Going silent since no one is currently listening to the channel").queue();
      event.getGuild().getAudioManager().closeAudioConnection();
      if (callback != null) callback.accept(event.getGuild().getIdLong());
      return;
    }

    playing.set(false);
    stop(event);
    play(event, commands);
    playing.set(true);
  }

  public void stop(GuildMessageReceivedEvent event) {
    player.stopTrack();
    voiceChannelId = null;
  }

  public void play(GuildMessageReceivedEvent event) throws Exception {
    play(event, null);
  }

  public void play(GuildMessageReceivedEvent event, List<String> commands) throws Exception {
    if (playing.get()) return;

    VoiceChannel vc = event.getMember().getVoiceState().getChannel();
    if (vc == null) {
      event.getChannel().sendMessage(String.format("<@%s> you are not in a voice channel to play music", event.getMember().getId())).queue();
      return;
    }

    if (!event.getGuild().getSelfMember().hasPermission(vc, Permission.VOICE_CONNECT, Permission.VOICE_SPEAK)) {
      event.getChannel().sendMessage(String.format("<@%s>, I cannot play my music in channel %s as I dont have permission to", 
          event.getMember().getId(), vc.getName())).queue();
      return;
    }

    Long tmpId = vc.getIdLong();
    if (voiceChannelId != null && voiceChannelId != tmpId) {
      String channelName = event.getGuild().getVoiceChannelById(voiceChannelId).getName();
      event.getChannel().sendMessage(String.format("<@%s> I am currently playing in channel %s. Cannot switch channels until I am stopped.", 
          event.getMember().getId(), channelName)).queue();
      return;
    }

    voiceChannelId = tmpId;

    AudioManager manager = event.getGuild().getAudioManager();
    manager.openAudioConnection(vc);

    manager.setSendingHandler(new EightTrackAudioSendHandler(player));

    if (commands != null && !commands.isEmpty()) {
      String uuid = commands.remove(0);
      boolean foundSong = findSong(uuid);
      if (!foundSong) {
        event.getChannel().sendMessageFormat("Unable to find song with identifier: %s", uuid).queue();
        return;
      }
      stop(event);
    }

    if (!ts.hasTracks()) {
      loadRandomSong();
    }

    if (ts.hasTracks()) {
      lastPlayed = OffsetDateTime.now(); 

      AudioTrack track = ts.getNextTrack();

      MessageEmbed embed = new EmbedBuilder()
          .setAuthor(event.getMember().getEffectiveName(), null, event.getAuthor().getAvatarUrl())
          .setTitle("Now Playing:")
          .setDescription(String.format("**%s** by __%s__", track.getInfo().title, track.getInfo().author))
          .setTimestamp(OffsetDateTime.now())
          .build();

      event.getChannel().sendMessage(embed).queue();
      player.playTrack(track);
    }
  }

  public void upload(GuildMessageReceivedEvent event) {
    if (!event.getMessage().getMember().hasPermission(Permission.ADMINISTRATOR, Permission.MANAGE_SERVER)) {
      event.getChannel().sendMessage(String.format("Sorry <@%s> i can't do that. *psst: you have to have manage server, or administrator role*", event.getMember().getIdLong())).queue();
      return;
    }

    for(Attachment attachment : event.getMessage().getAttachments()) {
      attachment.retrieveInputStream().thenAccept((in) -> {
        Mp3UploadResults res = null;
        MessageEmbed embed = null;
        try {

          res = uploadHandler.parse(in);
          if (res.validFile()) {
            embed = persistSong(res, event, attachment.getFileName());
          } else {
            embed = wrongFile(event, attachment.getFileName());
          }

        } catch (Exception ex) {
          logger.error("Unable to get file information: %s", ex.getMessage());
          logger.error(ex);
          embed = error(event, ex, attachment.getFileName()); 
        } finally {
          event.getMessage().delete().queue();
          if (res != null) res.clear();

          if (embed != null) event.getChannel().sendMessage(embed).queue();
        }

      }).exceptionally((err) -> {
        event.getMessage().delete().queue();
        MessageEmbed embed = error(event, err, attachment.getFileName());
        if (embed != null) event.getChannel().sendMessage(embed).queue();
        return null;
      });
    }
  }

  private MessageEmbed error(GuildMessageReceivedEvent event, Throwable ex, String fileName) {
    return new EmbedBuilder()
        .setTitle("Unable to upload file: " + fileName)
        .setDescription("There was an error uploading your file: " + ex.getMessage() + ". Please contact admin")
        .setColor(Color.RED)
        .setTimestamp(OffsetDateTime.now())
        .setFooter(event.getGuild().getSelfMember().getEffectiveName(), event.getGuild().getSelfMember().getUser().getAvatarUrl())
        .build();
  }

  private MessageEmbed wrongFile(GuildMessageReceivedEvent event, String fileName) {
    return new EmbedBuilder()
        .setTitle("Invalid File: " + fileName)
        .setDescription("Only music type files are allowed to be uploaded")
        .setColor(Color.RED)
        .setTimestamp(OffsetDateTime.now())
        .setFooter(event.getGuild().getSelfMember().getEffectiveName(), event.getGuild().getSelfMember().getUser().getAvatarUrl())
        .build();
  }

  private MessageEmbed persistSong(Mp3UploadResults result, GuildMessageReceivedEvent event, String fileName) throws Exception {

    DiscordGuild guild = guildSongRepoService.getGuildRepo().findByGuildId(event.getGuild().getIdLong());
    if(guild == null) {
      throw new IOException("Unable to find guild in database. Please contact administrator");
    }

    Path output = musicDir;
    if (result.getArtist() != null) {
      output = output.resolve(result.getArtist());
      if (result.getAlbum() != null) {
        output = output.resolve(result.getAlbum());
      }
    }
    output = output.resolve(fileName);

    if (!Files.exists(output.toAbsolutePath().getParent())) {
      Files.createDirectories(output.toAbsolutePath().getParent());
    }

    Files.write(output, result.getData());

    Song song = new Song();
    song.setAlbum(result.getAlbum());
    song.setArtist(result.getArtist());
    song.setDateAdded(new Date());
    song.setDiscNumber(result.getDiscNumber());
    if (result.getDuration() != null)
      song.setDuration(BigInteger.valueOf(result.getDuration()));
    song.setFileHash(HashUtils.hash("SHA-1", result.getData()));
    song.setFilePath(output.toAbsolutePath().toString());
    song.setGenre(result.getGenre());
    song.setReleaseYear(result.getReleaseDate());
    song.setTitle(result.getTitle());
    song.setTrackNumber(result.getTrackNumber());

    guildSongRepoService.getSongRepo().save(song);

    GuildSong gs = new GuildSong();
    gs.setDateAdded(new Date());
    gs.setGuild(guild);
    gs.setSong(song);

    String[] uuidArray = UUID.randomUUID().toString().split("-");

    gs.setUuid(uuidArray[uuidArray.length - 1]);

    guildSongRepoService.getGuildSongRepo().save(gs);

    MessageEmbed embed = new EmbedBuilder()
        .setColor(Color.GREEN)
        .setThumbnail(event.getGuild().getIconUrl())
        .setTitle("Upload Results for " + fileName)
        .addField("Title", result.getTitle(), true)
        .addField("Artist", result.getArtist(), true)
        .addField("Album", result.getAlbum(), true)
        .addField("Track", result.getTrackNumber(), true)
        .addField("Disc", result.getDiscNumber(), true)
        .addField("Year", result.getReleaseDate(), true)
        .addField("Duration", result.getDurationFormat(), true)
        .addField("ID", gs.getUuid(), true)
        .setTimestamp(OffsetDateTime.now())
        .setFooter(event.getGuild().getSelfMember().getEffectiveName(), event.getGuild().getSelfMember().getUser().getAvatarUrl())
        .build();

    return embed;
  } 

  public void list(GuildMessageReceivedEvent event) {
    List<GuildSong> gsList = guildSongRepoService.getGuildSongRepo().findByGuild(guildId);

    if (gsList == null || gsList.isEmpty()) {
      event.getChannel().sendMessage("There is no music for this guild.").queue();
      return;
    }

    int longestSong = 0;
    int longestArtist = 0;
    for(GuildSong gs : gsList) {
      if (gs.getSong().getTitle().length() > longestSong) longestSong = gs.getSong().getTitle().length();
      if (gs.getSong().getArtist().length() > longestArtist) longestArtist = gs.getSong().getArtist().length();
    }

    String fmt = "%6s | %-" + longestSong +"s | %-" + longestArtist +"s | %s%n"; 
    int count = 0;
    StringBuilder sb = new StringBuilder();
    sb.append("**" + "Currently available songs for " + event.getGuild().getName() + "**\n\n```");
    sb.append(String.format(fmt, "Track", "Title", "Artist", "id"));
    sb.append(StringUtils.repeat("-", 27 + longestSong + longestArtist)).append("\n");

    for(GuildSong gs : gsList) {
      sb.append(String.format(fmt, ++count, gs.getSong().getTitle(), gs.getSong().getArtist(), gs.getUuid()));
    }
    sb.append("```");
    event.getChannel().sendMessage(sb.toString()).queue();
  }

  private void loadRandomSong() throws Exception {
    List<GuildSong> gsList = guildSongRepoService.getGuildSongRepo().findByGuild(guildId);
    if (gsList == null || gsList.isEmpty()) return;

    Random random = new Random(System.currentTimeMillis());
    int item = random.nextInt(gsList.size());

    GuildSong song = gsList.get(item);

    apm.loadItem(song.getSong().getFilePath(), ts).get();
  }

  private boolean findSong(String uuid) throws Exception {
    GuildSong gs = guildSongRepoService.getGuildSongRepo().findByUuid(uuid);
    if (gs == null) return false;

    apm.loadItem(gs.getSong().getFilePath(), ts).get();

    return true;
  }

  public void playlist(GuildMessageReceivedEvent event, List<String> commands) {

    String command = commands.remove(0);

    switch(command) {
      case "add":
        addPlayList(event, commands);
        return;
      case "list":
        listPlayList(event, commands);
        return;
    }

  }

  private void listPlayList(GuildMessageReceivedEvent event, List<String> commands) {
    if (commands == null || commands.isEmpty()) {
      List<GuildPlaylist> userPlaylist = guildSongRepoService.getGuildPlaylistRepo().findByGuildAndUserIdFetchSongs(event.getGuild().getIdLong(), event.getMember().getIdLong());
      if (userPlaylist.isEmpty()) {
        event.getChannel().sendMessage(event.getMember().getAsMention() + " you have no defined playlists on this server").queue();
        return;
      }

      int longestPlaylist = 0;
      for(GuildPlaylist gpl : userPlaylist) {
        if (gpl.getPlaylist().length() > longestPlaylist) longestPlaylist = gpl.getPlaylist().length();
      }

      String fmt = "%6s | %-" + longestPlaylist +"s | %s%n"; 

      StringBuilder sb = new StringBuilder("```");
      sb.append(String.format(fmt, "", "Playlist", "Song Count"));
      sb.append(StringUtils.repeat("-", 18 + longestPlaylist)).append("\n");

      int count = 0;
      for(GuildPlaylist gpl : userPlaylist) {
        sb.append(String.format(fmt, ++count, gpl.getPlaylist(), gpl.getGuildPlaylistSongList().size()));
      }
      sb.append("```");
      event.getChannel().sendMessage("**" + "Currently available playlists for " + event.getMember().getAsMention() + "**").queue((succcess)-> 
        event.getChannel().sendMessage(sb.toString()).queue()
      );
      return;
    }
    
    String playlist = commands.remove(0);
    
    GuildPlaylist gpl = guildSongRepoService.getGuildPlaylistRepo().findGuildUserPlaylistFetchSongs(event.getGuild().getIdLong(), event.getMember().getIdLong(), playlist);
    if (gpl == null) {
      event.getChannel().sendMessage(event.getMember().getAsMention() + " you have no defined playlists on this server by the name of " + playlist).queue();
      return;
    }
    
    if (gpl.getGuildPlaylistSongList().isEmpty()) {
      event.getChannel().sendMessage(event.getMember().getAsMention() + " you have no defined songs for playlist " + playlist).queue();
      return;
    }
    
    int longestSong = 0;
    int longestArtist = 0;
    for(GuildPlaylistSong gpls : gpl.getGuildPlaylistSongList()) {
      GuildSong gs = gpls.getGuildSong();
      if (gs.getSong().getTitle().length() > longestSong) longestSong = gs.getSong().getTitle().length();
      if (gs.getSong().getArtist().length() > longestArtist) longestArtist = gs.getSong().getArtist().length();
    }

    String fmt = "%6s | %-" + longestSong +"s | %-" + longestArtist +"s | %s%n"; 
    int count = 0;
    StringBuilder sb = new StringBuilder("```");
    sb.append(String.format(fmt, "Track", "Title", "Artist", "id"));
    sb.append(StringUtils.repeat("-", 27 + longestSong + longestArtist)).append("\n");

    for(GuildPlaylistSong gpls : gpl.getGuildPlaylistSongList()) {
      GuildSong gs = gpls.getGuildSong();
      sb.append(String.format(fmt, ++count, gs.getSong().getTitle(), gs.getSong().getArtist(), gs.getUuid()));
    }
    sb.append("```");
    event.getChannel().sendMessage("**" + "Current songs in playlists " + playlist + " for " + event.getMember().getAsMention() + "**").queue((succcess)->
      event.getChannel().sendMessage(sb.toString()).queue()
    );

  }

  private void addPlayList(GuildMessageReceivedEvent event, List<String> commands) {
    if (commands == null || commands.isEmpty()) {
      event.getChannel().sendMessage(event.getAuthor().getAsMention() + " you have to provide the playlist to create a new playlist (no spaces in name) and optionally a list of uuids for songs to add").queue();
      return;
    }

    Long guildId = event.getGuild().getIdLong();

    DiscordGuild guild = guildSongRepoService.getGuildRepo().findByGuildId(guildId);
    if (guild == null) {
      event.getChannel().sendMessage("Unable to find guild in local database. Please contact administrator").queue();
      return;
    }

    String playlist = commands.remove(0);

    long userId = event.getMember().getIdLong();

    GuildPlaylist gpl = guildSongRepoService.getGuildPlaylistRepo().findByGuildAndUserIdAndPlaylist(guild, userId, playlist);
    boolean newList = false;
    if (gpl != null && commands.isEmpty()) {
      event.getChannel().sendMessage(event.getAuthor().getAsMention() + " a playlist with the name " + playlist + " already exist for you.").queue();
      return;
    } else if (gpl == null) {
      newList = true;
      gpl = new GuildPlaylist();
      gpl.setDateAdded(new Date());
      gpl.setGuild(guild);
      gpl.setPlaylist(playlist);
      gpl.setUserId(userId);
      guildSongRepoService.getGuildPlaylistRepo().save(gpl);
    }

    if (commands.isEmpty() && newList) {
      event.getChannel().sendMessage(event.getAuthor().getAsMention() + " playlist " + playlist + " successfully created.").queue();
      return;
    }

    Set<String> songIds = guildSongRepoService.getGuildPlaylistSongRepo()
        .findByGuildPlaylistAndSongIds(gpl, commands).stream()
        .map(g -> g.getGuildSong().getUuid())
        .collect(Collectors.toSet());

    List<GuildPlaylistSong> gplsList = new ArrayList<>();
    StringBuilder sb = new StringBuilder();
    for(Iterator<String> iter = commands.iterator(); iter.hasNext();) {
      String id = iter.next();
      GuildSong gs = guildSongRepoService.getGuildSongRepo().findByUuid(id);
      if (songIds.contains(id)) {
        sb.append(String.format("**%s** by __%s__ already exists in this playlist.%n", gs.getSong().getTitle(), gs.getSong().getArtist()));
        iter.remove();
        continue;
      }

      if (gs == null) {
        sb.append("Song with id `" + id + "` not found. Please check id\n");
        iter.remove();
        continue;
      }

      GuildPlaylistSong gpls = new GuildPlaylistSong();
      gpls.setDateAdded(new Date());
      gpls.setGuildPlaylist(gpl);
      gpls.setGuildSong(gs);
      gplsList.add(gpls);

      sb.append(String.format("**%s** by __%s__ added to playlist.%n", gs.getSong().getTitle(), gs.getSong().getArtist()));
      iter.remove();
    }

    if (!gplsList.isEmpty()) {
      guildSongRepoService.getGuildPlaylistSongRepo().saveAll(gplsList);
    }

    MessageEmbed embed = new EmbedBuilder()
        .setTitle("Results for adding playlist " + playlist)
        .setAuthor(event.getMember().getEffectiveName(), null, event.getAuthor().getAvatarUrl())
        .setColor(Color.GREEN)
        .setDescription(sb.toString())
        .setFooter("", event.getGuild().getSelfMember().getUser().getAvatarUrl())
        .setTimestamp(OffsetDateTime.now())
        .build();

    event.getChannel().sendMessage(embed).queue();

  }
}
